/*
 * Rutinas para capturar imgenes en Video4Linux
 * =============================================
 * 
 * (c) 1998-2002 Daniel Serpell C.
 * 
 */
 
/*
 * capture.c: Rutinas para capturar cuadros a memoria.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <linux/types.h>
#include <linux/videodev.h>

#include "capture.h"

/* Datos privados. Estas rutinas no son reentrantes... */
struct _struct_capture
{
#define NUM_BUFFERS (2)
   int vid;
   int sizex;
   int sizey;
   int bufsize;
   unsigned int format;
   unsigned char *vbuf[NUM_BUFFERS];
};

/* Abre el dispositivo, pide las capacidades y setea los parmetros */
struct_capture *capture_init(struct capture_options *copt, int verbose)
{
    int sx,sy;
    int vid, i;
    static struct video_capability vcap;
    static struct video_mbuf vmbuf;
    static struct video_channel vchan;
    static struct video_picture vpict;
    struct_capture *data;
    unsigned char *vbuf[NUM_BUFFERS];
   
    vid = open( copt->devname, O_RDONLY );
    if ( vid == -1 )
    {
        fprintf(stderr,"capture.init(): Error al abrir el dispositivo %s, %s\n",
	      copt->devname, strerror(errno));
        return 0;
    }
    if(ioctl( vid, VIDIOCGCAP, &vcap ))
    {
        fprintf(stderr,"capture.init(): Error al pedir capacidades del disp, %s\n",
	      strerror(errno));
        close(vid);
        return 0;
    }
    if( !(vcap.type & VID_TYPE_CAPTURE) )
    {
	fprintf(stderr,"capture.init(): Error, el dispositivo no puede capturar.\n");
	close(vid);
	return 0;
    }
    vchan.channel = copt->channel;
    vchan.norm = VIDEO_MODE_NTSC;

    if( ioctl( vid, VIDIOCSCHAN, &vchan ))
    {
    	fprintf(stderr,"Error al seleccionar canal, %s\n", strerror(errno));
        close(vid);
        return 0;
    }   

    if(ioctl( vid, VIDIOCGCHAN, &vchan ))
    {
        fprintf(stderr,"Error al buscar canales, %s\n", strerror(errno));
        close(vid);
        return 0;
    }   

    vpict.brightness = 30000;
    vpict.contrast = 30000;
    vpict.whiteness = 1;
    vpict.palette = copt->format;

    if( ioctl( vid, VIDIOCSPICT, &vpict ))
    {
    	fprintf(stderr,"Error al setear paleta, %s\n", strerror(errno));
        close(vid);
        return 0;
    }   

    if( ioctl( vid, VIDIOCGPICT, &vpict) )
    {
    	fprintf(stderr,"Error en los parametros de la imagen, %s\n", strerror(errno));
        close(vid);
        return 0;
    }   
   
    /* Verifica el tamao del rectngulo a capturar */
    sx = copt->sizex; sy = copt->sizey;
   
    if( vcap.type & VID_TYPE_SUBCAPTURE )
    {
        /* Slo captura un rectngulo de la imgen  (sx*sy) */
        struct video_capture vct;
      
        if (vcap.maxwidth  < sx) sx = vcap.maxwidth;
        if (vcap.maxheight < sy) sy = vcap.maxheight;

        vct.x = (vcap.maxwidth  - sx) / 2;
        vct.y = (vcap.maxheight - sy) / 2;
        vct.width  = sx;
        vct.height = sy;
      
        vct.decimation = 0;
        vct.flags = 0;
      
        if(ioctl( vid, VIDIOCSCAPTURE, & vct ))
        {
	    fprintf(stderr,"capture.init(): Error al setear parmetros de sub-captura, %s\n",
		 strerror(errno));
	    close(vid);
	    return 0;
        }
    }
    else if ( vcap.type & VID_TYPE_SCALES )
    {
        if (vcap.maxwidth  < sx) sx = vcap.maxwidth;
        if (vcap.maxheight < sy) sy = vcap.maxheight;
    }
    else
    {
        sx = vcap.maxwidth;
        sy = vcap.maxheight;
    }
   
    /* Hace MMAP */
    if(ioctl( vid, VIDIOCGMBUF, &vmbuf ))
    {
        fprintf(stderr,"capture.init(): Error al pedir buffer para mmap, %s\n",
	      strerror(errno));
        close(vid);
        return 0;
    }
    
    if( vmbuf.frames < NUM_BUFFERS)
    {
        fprintf(stderr,"capture.init(): "
		"Error, necesito un buffer de al menos %d cuadros.\n", NUM_BUFFERS);
        close(vid);
        return 0;
    }

    for(i=0;i<NUM_BUFFERS;i++)
    {
	vbuf[i] = mmap(0, sx*sy, PROT_READ /*|PROT_WRITE*/ , MAP_SHARED, vid,
		vmbuf.offsets[i] );
	if( (int)vbuf[i] == -1)
	{
	    fprintf(stderr,"capture.init(): Error al realizar mmap, %s\n",
		    strerror(errno));
	    close(vid);
	    return 0;
	}
    }

    if(verbose)
       printf("Capacidades del dispositivo:\n"
	      "\tNombre:%s\n"
	      "\tNmero de canales de video:%d\n"
	      "\tNmero de canales de audio:%d\n"
	      "\tMximo tamao de captura:%dx%d\n"
	      "\tMnimo tamao de captura:%dx%d\n"
	      "\tFlags:\t%s%s%s%s%s%s%s%s%s%s\n",
	      vcap.name, vcap.channels, vcap.audios, 
	      vcap.maxwidth, vcap.maxheight, vcap.minwidth, vcap.minheight,
	      (vcap.type & VID_TYPE_CAPTURE) ? "Captura a memoria.\n\t\t" : "",
	      (vcap.type & VID_TYPE_TUNER)   ? "Posee sintonisador.\n\t\t" : "",
	      (vcap.type & VID_TYPE_TELETEXT) ? "Lee teletexto.\n\t\t" : "",
	      (vcap.type & VID_TYPE_OVERLAY) ? "Puede sobreponer imgen en pantalla.\n\t\t" : "",
	      (vcap.type & VID_TYPE_CHROMAKEY) ? "Sobreposicin utiliza clave de color.\n\t\t" : "",
	      (vcap.type & VID_TYPE_CLIPPING) ? "Soporta recorte de la superposicin.\n\t\t" : "",
	      (vcap.type & VID_TYPE_FRAMERAM) ? "Superposicin sobreescribe mem. video.\n\t\t" : "",
	      (vcap.type & VID_TYPE_SCALES) ? "Soporta escalamiento por hardware.\n\t\t" : "",
	      (vcap.type & VID_TYPE_MONOCHROME) ? "Captura slo en grises.\n\t\t" : "",
	      (vcap.type & VID_TYPE_SUBCAPTURE) ? "Puede capturar subconjunto de la imgen.\n\t\t" : "\n" );

    if(verbose)
    {
	printf("\nCaracteristica del canal:\n"
	      "\tCanal:  %d\n"
	      "\tNombre: %s\n"
	      "\tFlags:  %s%s"
	      "\tTipo:   %s%s"
	      "\tNorma:  %s%s%s%s",
	      vchan.channel, vchan.name, 
	      (vchan.flags & VIDEO_VC_TUNER) ? "Canal tiene sintonizador.\n\t\t" : "",
	      (vchan.flags & VIDEO_VC_AUDIO)   ? "Canal tiene audio.\n" : "",
	      (vchan.type & VIDEO_TYPE_TV) ? "Entrada de Television.\n" : "",
	      (vchan.type & VIDEO_TYPE_CAMERA) ? "Entrada de camara.\n" : "",	
	      (vchan.norm & VIDEO_MODE_PAL) ? "PAL.\n\t\t" : "",
	      (vchan.norm & VIDEO_MODE_NTSC) ? "NTSC.\n\t\t" : "",	
	      (vchan.norm & VIDEO_MODE_SECAM) ? "SECAM.\n\t\t" : "",
	      (vchan.norm & VIDEO_MODE_AUTO) ? "AUTO.\n\t\t" : "\n");

	printf("\nParametros de la Imagen:\n"
	      "\tBrillo:    %d\n"
	      "\tHue:       %d\n"
	      "\tColour:    %d\n"
	      "\tContrast:  %d\n"
	      "\tWhiteness: %d\n"
	      "\tDepth:     %d\n"
	      "\tPaleta:    %s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s\n",
	      vpict.brightness, vpict.hue, vpict.colour, vpict.contrast,
	      vpict.whiteness, vpict.depth, 
	      (vpict.palette & VIDEO_PALETTE_GREY) ? "Escala de grises\n\t\t" : "", 
	      (vpict.palette & VIDEO_PALETTE_HI240) ? "HI240\n\t\t" : "", 
	      (vpict.palette & VIDEO_PALETTE_RGB565) ? "   RGB565\n\t\t" : "", 
	      (vpict.palette & VIDEO_PALETTE_RGB24) ? "   RGB24\n\t\t" : "", 
	      (vpict.palette & VIDEO_PALETTE_RGB32) ? "   RGB32\n\t\t" : "", 
	      (vpict.palette & VIDEO_PALETTE_RGB555) ? "   RGB555\n" : "", 	      
	      (vpict.palette & VIDEO_PALETTE_YUV422) ? "YUV422\n" : "", 
	      (vpict.palette & VIDEO_PALETTE_YUYV) ? "YUYV\n" : "", 
	      (vpict.palette & VIDEO_PALETTE_UYVY) ? "UYVY\n" : "", 
	      (vpict.palette & VIDEO_PALETTE_YUV420) ? "YUV420\n" : "", 
	      (vpict.palette & VIDEO_PALETTE_YUV411) ? "YUV411\n" : "", 
	      (vpict.palette & VIDEO_PALETTE_RAW) ? "RAW\n" : "", 
	      (vpict.palette & VIDEO_PALETTE_YUV422P) ? "YUV422P\n" : "", 
	      (vpict.palette & VIDEO_PALETTE_YUV411P) ? "YUV411P\n" : "", 
	      (vpict.palette & VIDEO_PALETTE_YUV420P) ? "YUV420P\n" : "", 
	      (vpict.palette & VIDEO_PALETTE_YUV410P) ? "YUV410P\n" : "\n" );

    }
    if(verbose)
    {
        int j;
        printf("Capacidades de MMAP:\n"
	     "\tTamao:      0x%X bytes\n"
	     "\tCuadros:     %d\n", vmbuf.size, vmbuf.frames );
        for(j=0;j<vmbuf.frames;j++)
	  printf("\tOffset[%d] = 0x%X\n\n", j,vmbuf.offsets[j]);
    }
   
    /* Arregla los punteros */
    data = malloc(sizeof(struct_capture));
    data->vbuf[0] = vbuf[0];
    data->vbuf[1] = vbuf[1];
    data->vid = vid;
    data->sizex = sx;
    data->sizey = sy;
    data->bufsize = vmbuf.size;
    data->format = copt->format;
    
    return data;
}

unsigned char * capture_getbuffer(struct_capture *data, int f)
{
    if((f<0)||(f>1)||(!data))
	return 0;
    return data->vbuf[f];
}

int capture_getbufsize(struct_capture *data)
{
    if (data)
	return data->bufsize;
    else
       return 0;
}

int capture_getframe(struct_capture *data, int fnum )
{
   struct video_mmap vmmap;
   if(!data)
      return 0;

   /* Comienza a capturar un cuadro*/
   vmmap.frame = fnum;
   vmmap.height = data->sizey;
   vmmap.width  = data->sizex;
   vmmap.format = data->format;
   if(ioctl( data->vid, VIDIOCMCAPTURE, &vmmap ))
   {
      fprintf(stderr, "capture.getframe(): Error al capturar cuadro %d, %s\n", fnum, strerror(errno));
      return 0;
   }
   return 1;
}

int capture_syncframe(struct_capture *data, int fnum )
{
   if(!data)
      return 0;
   
   if(ioctl( data->vid, VIDIOCSYNC, &fnum ))
   {
      fprintf(stderr, "capture.freeframe(): Error al sincronizar con el cuadro %d, %s\n", fnum, strerror(errno));
      return 0;
   }

   return 1;
}

